from openride.core.numba.interpolation import interpolate_line_index_at_xy, linear_interpolation

import numba
import numpy as np

from typing import Tuple, Union

EPSILON = np.finfo(float).eps


@numba.njit(cache=True)
def vertices_to_distances(vertices:np.ndarray, bird_eye_view:bool=True) -> np.ndarray:
    distances_squared = (vertices[1:,0] - vertices[:-1,0]) ** 2 + (vertices[1:,1] - vertices[:-1,1]) ** 2
    if not bird_eye_view:
        distances_squared += (vertices[1:,2] - vertices[:-1,2]) ** 2
    distances = distances_squared ** 0.5
    distances = np.append(0, distances)
    return np.cumsum(distances)


@numba.njit(cache=True)
def index_to_distance(vertices:np.ndarray, index:float) -> float:
    distances = vertices_to_distances(vertices)
    if index >= vertices.shape[0]-1: return distances[-1]
    return linear_interpolation(distances, index)


@numba.njit(cache=True)
def distance_to_index(vertices:np.ndarray, distance:float) -> float:
    if distance <= 0.0: return 0.0
    distances = vertices_to_distances(vertices)
    index = np.searchsorted(distances - distance, 0)
    if index > distances.size - 1: return distances.size - 1
    index += (distance - distances[index])/(distances[index] - distances[index-1] + EPSILON)
    return index


@numba.njit(cache=True)
def left_or_right_from_line(vertices:np.ndarray, x:float, y:float) -> int:
    i = int(interpolate_line_index_at_xy(vertices, x, y))
    i0 = max([i, 0])
    i1 = min([i + 1, vertices.shape[0] - 1])
    return np.sign((vertices[i1,0] - vertices[i0,0]) * (y - vertices[i0,1]) - (vertices[i1,1] - vertices[i0,1]) * (x - vertices[i0,0]))


@numba.njit(cache=True)
def tangent_angle(vertices:np.ndarray, index:int) -> float:
    i1 = max([0, index-1])
    i2 = min([index+1, vertices.shape[0]-1])
    delta_x = vertices[i2,0] - vertices[i1,0]
    delta_y = vertices[i2,1] - vertices[i1,1]
    ratio = delta_y / (delta_x + 1e-9)
    angle = np.arctan(ratio)
    angle = angle + np.pi if delta_x < 0 else angle
    return angle % (2*np.pi)


@numba.njit(cache=True)
def tangent_angles(vertices:np.ndarray) -> np.ndarray:
    angles = np.zeros(vertices.shape[0])
    delta_x = vertices[2:,0] - vertices[:-2,0]
    delta_y = vertices[2:,1] - vertices[:-2,1]
    ratio = delta_y / (delta_x + 1e-9)
    angles[1:-1] = np.arctan(ratio)
    angles[1:-1][np.where(delta_x < 0)] += np.pi
    angles[1:-1] = angles[1:-1] % (2*np.pi)
    angles[0] = tangent_angle(vertices, 0)
    angles[-1] = tangent_angle(vertices, vertices.shape[0]-1)
    return angles


@numba.njit(cache=True)
def angle_difference(angle1:float, angle2:float) -> float:
    two_pi = 2.0 * np.pi
    return min((two_pi - abs(angle1 - angle2), abs(angle1 - angle2))) % two_pi

@numba.njit(cache=True)
def angle_differences(angles1:float, angles2:float) -> float:
    two_pi = 2.0 * np.pi
    return np.array([min((two_pi - abs(angles1[i] - angles2[i]), abs(angles1[i] - angles2[i]))) % two_pi for i in range(angles1.size)])


@numba.njit(cache=True)
def project_point_on_line(vertices:np.ndarray, x:float, y:float) -> Tuple[float, float, float]:
    """Projection is performed in bird eye view, but the coordinate interpolation is 3D."""
    index = interpolate_line_index_at_xy(vertices, x, y)
    vx = linear_interpolation(vertices[:,0], index)
    vy = linear_interpolation(vertices[:,1], index)
    vz = linear_interpolation(vertices[:,2], index)
    return vx, vy, vz


@numba.njit(cache=True)
def radius_of_curvature(vertices: np.ndarray) -> np.ndarray:
    # https://stackoverflow.com/questions/61945776/curve-radius-in-python
    R = np.full(vertices.shape[0], np.inf)

    dx = vertices[1:,0] - vertices[:-1,0] # First derivative of x (centered)
    dy = vertices[1:,1] - vertices[:-1,1] # First derivative of y (centered)
    ddx = dx[1:] - dx[:-1] # Second derivative of x (centered)
    ddy = dy[1:] - dy[:-1] # Second derivative of y (centered)

    # Remove edges to match shapes
    dx = dx[:-1]
    dy = dy[:-1]

    R[:-2] = (dx**2 + dy**2)**1.5 / (dx*ddy - dy*ddx + EPSILON)

    # Deal with edges
    R[-2:] = R[-3]

    return R


@numba.njit(cache=True)
def apply_lateral_deviations(vertices:np.ndarray, tangent_angles:np.ndarray, lateral_deviations:Union[np.ndarray, float]) -> np.ndarray:
    """Returns the route vertices after applying the lateral deviations"""
    dx = lateral_deviations * np.sin(tangent_angles)
    dy = lateral_deviations * np.cos(tangent_angles)
    shifted_vertices = np.copy(vertices)
    shifted_vertices[:,0] -= dx
    shifted_vertices[:,1] += dy
    return shifted_vertices


@numba.njit(cache=True)
def resample_linear(vertices:np.ndarray, N:int):
    distances = vertices_to_distances(vertices)
    new_distances = np.linspace(0, distances[-1], N)
    new_vertices = np.zeros((new_distances.size, 3))
    indices = np.zeros(new_distances.size)

    for i in range(new_distances.size):
        indices[i] = distance_to_index(vertices, new_distances[i])
        new_vertices[i,0] = linear_interpolation(vertices[:,0], indices[i])
        new_vertices[i,1] = linear_interpolation(vertices[:,1], indices[i])
        new_vertices[i,2] = linear_interpolation(vertices[:,2], indices[i])

    return new_vertices, indices



import openride as ride

from dataclasses import dataclass
from typing import Optional

import time
import numpy as np



@dataclass
class Hero:
    bounding_box:ride.BoundingBox = ride.BoundingBox(ride.Point(0,0,0.8), rotation=ride.Rotation(), size=ride.Size(2.3,0.9,0.8))
    max_velocity:float = 50.0

    def __post_init__(self):
        self.velocity:float = 0.0
        self.acceleration:float = 0.0
        self.steering:float = 0.0


    def update(self, dt:float):

        steering = self.steering / abs(self.velocity + 1)

        x, y, yaw, v = rear_axle_kinematic_model(
            self.bounding_box.position.x, self.bounding_box.position.y, self.bounding_box.rotation.yaw,
            self.velocity, dt, self.acceleration, steering, self.bounding_box.size.x*2, 0.5,
        )

        self.bounding_box.position.x = x
        self.bounding_box.position.y = y
        self.bounding_box.rotation.yaw = yaw
        self.velocity = min((v, self.max_velocity))


    def on_key_press(self, key_event:ride.KeyEvent):

        if key_event.key == 'w' or key_event.key == 'Up':
            self.acceleration = 25.0
        if key_event.key == 's' or key_event.key == 'Down':
            self.acceleration = -25.0

        if key_event.key == 'a' or key_event.key == 'Left':
            self.steering = 4.0
        if key_event.key == 'd' or key_event.key == 'Right':
            self.steering = -4.0


    def on_key_release(self, key_event:ride.KeyEvent):
        if key_event.key in ['w','s','Up','Down']:
            self.acceleration = 0.0
        if key_event.key in ['a','d','Left','Right']:
            self.steering = 0.0


    def collides(self, colliders:ride.BoundingBoxCollection) -> Optional[int]:
        for index, collider in enumerate(colliders):
            if ride.bird_eye_view_distance(self.bounding_box, collider) == 0.0:
                return index


    def respawn(self):
        self.bounding_box = ride.BoundingBox(ride.Point(0,0,0.8), rotation=ride.Rotation(), size=ride.Size(2.3,0.9,0.8))
        self.velocity = 0.0
        self.acceleration = 0.0
        self.steering = 0.0


def rear_axle_kinematic_model(x0, y0, yaw0, v0, dt, acceleration, wheel_angle, wheelbase, drag):
    x = x0 + (v0 * np.cos(yaw0) * dt)
    y = y0 + (v0 * np.sin(yaw0) * dt)
    v = v0 + (acceleration * dt)
    v -= v0 * drag * dt
    yaw = yaw0 + (v0 * np.tan(wheel_angle) / wheelbase * dt)
    return x, y, yaw, v


@dataclass
class City:
    size:float = 200
    building_width:float = 25
    road_width:float = 15

    def __post_init__(self):
        self.buildings = ride.BoundingBoxCollection()
        self.traffic_cones = ride.BoundingBoxCollection()
        self.generate_buildings()
        self.generate_traffic_cones()


    def generate_buildings(self, fill_ratio:float=0.9):
        for x in range(-self.size, self.size, self.building_width+self.road_width):
            for y in range(-self.size, self.size, self.building_width+self.road_width):
                if np.random.random() > fill_ratio: continue
                height = np.random.random()*15+5
                building = ride.BoundingBox(
                    ride.Point(x,y,height), 
                    size=ride.Size(self.building_width/2,self.building_width/2,height)
                )
                self.buildings.append(building)
        offset = self.size % (self.building_width + self.road_width) + (self.building_width + self.road_width)/2
        self.buildings = self.buildings.transform(ride.Transform(ride.Point(0, offset)))


    def spawn_traffic_cone(self):
        x = np.random.random()*2*self.size - self.size
        y = np.random.random()*2*self.size - self.size
        box = ride.BoundingBox(ride.Point(x, y, 0.5), size=ride.Size(0.5,0.5,0.5))
        for building in self.buildings:
            if ride.bird_eye_view_distance(box, building) == 0.0:
                self.spawn_traffic_cone()
        self.traffic_cones.append(box)


    def generate_traffic_cones(self, density:float=0.01):
        amount = int(density * self.size ** 2)
        for _ in range(amount):
            self.spawn_traffic_cone()


def main():

    hero = Hero()
    city = City(size=220)

    viewer = ride.Viewer(mouse_camera_interactions=False)
    viewer.draw_grid(extent=city.size+40)
    viewer.hud.toggle_fps_count()
    viewer.interactor.set_callback(ride.Event.KeyPressEvent, hero.on_key_press)
    viewer.interactor.set_callback(ride.Event.KeyReleaseEvent, hero.on_key_release)

    score = 0

    t = time.time()
    while True:
        dt = time.time() - t
        t = time.time()

        if hero.collides(city.buildings):
            hero.respawn()
            viewer.camera.follow(hero.bounding_box, spring=0)
            score -= 5

        collision_index = hero.collides(city.traffic_cones)
        if collision_index:
            city.traffic_cones.pop(collision_index)
            city.spawn_traffic_cone()
            score += 1

        score = max((score, 0))

        hero.update(dt)
        viewer.camera.follow(hero.bounding_box, spring=0.8)

        viewer.draw_model(ride.SubCategory.Car, hero.bounding_box, color=(0.7,0.6,0.35))
        [viewer.draw_bounding_box(building, color=(0.5, 0.5, 0.6)) for building in city.buildings if viewer.camera.is_in_front(building.position)]
        [viewer.draw_model(ride.Category.TrafficCone, cone, color=(0.9, 0.2, 0.1), number_polygons=200) for cone in city.traffic_cones if viewer.camera.is_in_front(cone.position)]

        viewer.hud.add('score', score)
        viewer.update()


if __name__ == "__main__":
    main()


from openride.core.numba.interpolation import interpolate_line_index_at_xy, linear_interpolation
from openride.core.numba.polygon_utils import polygon_contains_point
from openride.core.numba.box_utils import bev_box_vertices

from typing import Tuple

import numba
import numpy as np



@numba.njit(cache=True)
def bev_distance_between_points(x1:float, y1:float, x2:float, y2:float) -> float:
    return ((x1-x2)**2 + (y1-y2)**2)**0.5



@numba.njit(cache=True)
def bev_distance_between_box_and_point(box_position_x:float, box_position_y:float, box_size_x:float, box_size_y:float, box_yaw:float, point_x:float, point_y:float) -> float:
    relx = point_x - box_position_x
    rely = point_y - box_position_y
    rotx = relx*np.cos(-box_yaw) - rely*np.sin(-box_yaw)
    roty = relx*np.sin(-box_yaw) + rely*np.cos(-box_yaw)
    dx = max(abs(rotx) - box_size_x, 0)
    dy = max(abs(roty) - box_size_y, 0)
    return (dx * dx + dy * dy)**0.5


@numba.njit(cache=True)
def bev_project_point_on_line(vertices:np.ndarray, x:float, y:float) -> Tuple[float, float]:
    index = interpolate_line_index_at_xy(vertices, x, y)
    vx = linear_interpolation(vertices[:,0], index)
    vy = linear_interpolation(vertices[:,1], index)
    return vx, vy


@numba.njit(cache=True)
def bev_distance_between_line_and_point(vertices:np.ndarray, x:float, y:float) -> float:
    return min([point_segment_distance(x, y, vertices[i,0], vertices[i,1], vertices[i+1,0], vertices[i+1,1]) for i in range(vertices.shape[0]-1)])


@numba.njit(cache=True)
def bev_distance_between_polygon_and_point(vertices:np.ndarray, x:float, y:float) -> float:
    if polygon_contains_point(vertices, x, y): return 0.0
    return bev_distance_between_line_and_point(np.vstack((vertices, vertices[:0])), x, y)


@numba.njit(cache=True)
def bev_distance_between_line_and_polygon(line_vertices:np.ndarray, poly_vertices:np.ndarray) -> float:
    N1 = line_vertices.shape[0]
    N2 = poly_vertices.shape[0]
    
    distances = np.zeros((N1-1,N2))
    for i1 in range(N1-1):
        x11, y11 = line_vertices[i1,:2]
        x12, y12 = line_vertices[(i1+1),:2]

        if polygon_contains_point(poly_vertices, x11, y11): return 0.0

        for i2 in range(N2):
            x21, y21 = poly_vertices[i2%N2,:2]
            x22, y22 = poly_vertices[(i2+1)%N2,:2]

            distances[i1,i2] = segments_distance(x11, y11, x12, y12, x21, y21, x22, y22)
            if distances[i1,i2] == 0.0: return 0.0

    return np.min(distances)


@numba.njit(cache=True)
def bev_distance_between_line_and_box(line_vertices:np.ndarray, x, y, sx, sy, yaw) -> float:
    return bev_distance_between_line_and_polygon(line_vertices, bev_box_vertices(x, y, sx, sy, yaw))


@numba.njit(cache=True)
def bev_distance_between_polygon_and_box(poly_vertices:np.ndarray, x, y, sx, sy, yaw) -> float:
    return bev_distance_between_polygons(poly_vertices, bev_box_vertices(x, y, sx, sy, yaw))


@numba.njit(cache=True)
def bev_distance_between_polygons(vertices1:np.ndarray, vertices2:np.ndarray) -> float:

    for x, y in vertices1[:,:2]:
        if polygon_contains_point(vertices2, x, y): return 0.0
    for x, y in vertices2[:,:2]:
        if polygon_contains_point(vertices1, x, y): return 0.0

    N1 = vertices1.shape[0]
    N2 = vertices2.shape[0]
    
    distances = np.zeros((N1,N2))
    for i1 in range(N1):
        x11, y11 = vertices1[i1%N1,:2]
        x12, y12 = vertices1[(i1+1)%N1,:2]

        for i2 in range(N2):
            x21, y21 = vertices2[i2%N2,:2]
            x22, y22 = vertices2[(i2+1)%N2,:2]

            distances[i1,i2] = segments_distance(x11, y11, x12, y12, x21, y21, x22, y22)
            if distances[i1,i2] == 0.0: return 0.0

    return np.min(distances)


@numba.njit(cache=True)
def bev_distance_between_boxes(x1, y1, sx1, sy1, yaw1, x2, y2, sx2, sy2, yaw2) -> float:
    return bev_distance_between_polygons(
        bev_box_vertices(x1, y1, sx1, sy1, yaw1), 
        bev_box_vertices(x2, y2, sx2, sy2, yaw2),
    )


@numba.njit(cache=True)
def bev_distance_between_lines(vertices1:np.ndarray, vertices2:np.ndarray) -> float:
    N1 = vertices1.shape[0]
    N2 = vertices2.shape[0]

    if N1 == 1 and N2 == 1:
        return bev_distance_between_points(vertices1[0,0], vertices1[0,1], vertices1[1,0], vertices1[1,1])
    elif N1 == 1:
        return bev_distance_between_line_and_point(vertices2, vertices1[0,0], vertices1[0,1])
    elif N2 == 1:
        return bev_distance_between_line_and_point(vertices1, vertices2[1,0], vertices2[1,1])
    
    distances = np.zeros((N1-1, N2-1))
    for i1 in range(N1-1):
        x11, y11 = vertices1[i1,:2]
        x12, y12 = vertices1[(i1+1),:2]

        for i2 in range(N2-1):
            x21, y21 = vertices2[i2,:2]
            x22, y22 = vertices2[(i2+1),:2]

            distances[i1,i2] = segments_distance(x11, y11, x12, y12, x21, y21, x22, y22)
            if distances[i1,i2] == 0.0: return 0.0

    return np.min(distances)


@numba.njit(cache=True)
def segments_distance(x11, y11, x12, y12, x21, y21, x22, y22):
    """ distance between two segments in the plane:
        one segment is (x11, y11) to (x12, y12)
        the other is   (x21, y21) to (x22, y22)
    """
    if segments_intersect(x11, y11, x12, y12, x21, y21, x22, y22): return 0.0
    # try each of the 4 vertices w/the other segment
    distances = (
        point_segment_distance(x11, y11, x21, y21, x22, y22),
        point_segment_distance(x12, y12, x21, y21, x22, y22),
        point_segment_distance(x21, y21, x11, y11, x12, y12),
        point_segment_distance(x22, y22, x11, y11, x12, y12),
    )
    return min(distances)


@numba.njit(cache=True)
def segments_intersect(x11, y11, x12, y12, x21, y21, x22, y22):
    """ whether two segments in the plane intersect:
        one segment is (x11, y11) to (x12, y12)
        the other is   (x21, y21) to (x22, y22)
    """
    dx1 = x12 - x11
    dy1 = y12 - y11
    dx2 = x22 - x21
    dy2 = y22 - y21
    delta = dx2 * dy1 - dy2 * dx1
    if delta == 0: return False  # parallel segments
    s = (dx1 * (y21 - y11) + dy1 * (x11 - x21)) / delta
    t = (dx2 * (y11 - y21) + dy2 * (x21 - x11)) / (-delta)
    return (0 <= s <= 1) and (0 <= t <= 1)


@numba.njit(cache=True)
def point_segment_distance(px, py, x1, y1, x2, y2):
    dx = x2 - x1
    dy = y2 - y1
    if dx == dy == 0:  # the segment's just a point
        return ((px-x1)**2 + (py-y1)**2)**0.5 

    # Calculate the t that minimizes the distance.
    t = ((px - x1) * dx + (py - y1) * dy) / (dx * dx + dy * dy)

    # See if this represents one of the segment's
    # end points or a point in the middle.
    if t < 0:
        dx = px - x1
        dy = py - y1
    elif t > 1:
        dx = px - x2
        dy = py - y2
    else:
        near_x = x1 + t * dx
        near_y = y1 + t * dy
        dx = px - near_x
        dy = py - near_y

    return (dx**2+dy**2)**0.5
    


from openride.core.geometry import Geometry
from openride.core.numba.transforms import transform_vertices
from openride.core.point import Point

import copy
import numpy as np

from dataclasses import dataclass
from shapely import geometry
from typing import List, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from openride.core.transform import Transform



@dataclass
class Polygon(Geometry):

    vertices: Union[np.ndarray, List[Point]] = np.empty((0,3))


    def __post_init__(self):

        if isinstance(self.vertices, list):
            points = copy.deepcopy(self.vertices)
            self.vertices = np.zeros((len(points), 3))
            for i, point in enumerate(points):
                self.vertices[i] = point.x, point.y, point.z

        if np.any(self.vertices[0] != self.vertices[-1]):
            self.vertices = np.vstack([self.vertices, self.vertices[0]])


    def to_shapely(self) -> geometry.Polygon:
        return geometry.Polygon(self.vertices)


    def transform(self, transform:'Transform') -> 'Polygon':
        return Polygon(transform_vertices(self.vertices, transform.get_matrix()))


    @classmethod
    def from_shapely(cls, polygon:geometry.Polygon) -> 'Polygon':
        return cls([Point(*c) for c in polygon.exterior.coords])
    

    def get_area(self) -> float:
        return self.to_shapely().area



from openride import Point, BoundingBox, Polyline, Polygon, Size, bird_eye_view_contains
from openride.test import get_pytest_benchmark



pytest_benchmark = get_pytest_benchmark(group="operations.bird_eye_view_contains")


@pytest_benchmark
def test_bev_contains_box_point_true(benchmark):
    box = BoundingBox(Point(4,8,-3))
    point = Point(4.5,7.5,5)
    assert benchmark(bird_eye_view_contains, box, point)


def test_bev_contains_box_point_false():
    assert not bird_eye_view_contains(BoundingBox(), Point(0,5,0))


@pytest_benchmark
def test_bev_contains_box_box_true(benchmark):
    box1 = BoundingBox()
    box2 = BoundingBox(Point(0,0,5), size=Size(0.5,0.5,2.5))
    assert benchmark(bird_eye_view_contains, box1, box2)


def test_bev_contains_box_box_false():
    assert not bird_eye_view_contains(BoundingBox(), BoundingBox(Point(0,0,5), size=Size(2.5,0.5,2.5)))


@pytest_benchmark
def test_bev_contains_box_polyline_true(benchmark):
    box = BoundingBox()
    polyline = Polyline([Point(), Point(0.5,0.3,2), Point(-0.3,0.7,-2)])
    assert benchmark(bird_eye_view_contains, box, polyline)


def test_bev_contains_box_polyline_false():
    polyline = Polyline([Point(), Point(2.5,0.3,2), Point(-0.3,0.7,-2)])
    assert not bird_eye_view_contains(BoundingBox(), polyline)


@pytest_benchmark
def test_bev_contains_box_polygon_true(benchmark):
    box = BoundingBox()
    polygon = Polygon([Point(), Point(0.5,0.3,2), Point(-0.3,0.7,-2)])
    assert benchmark(bird_eye_view_contains, box, polygon)


def test_bev_contains_box_polygon_false():
    polygon = Polygon([Point(), Point(2.5,0.3,2), Point(-0.3,0.7,-2)])
    assert not bird_eye_view_contains(BoundingBox(), polygon)


@pytest_benchmark
def test_bev_contains_polyline_point_true(benchmark):
    polyline = Polyline([Point(x,x,x) for x in range(5)])
    point = Point(1.3,1.3,-5)
    assert benchmark(bird_eye_view_contains, polyline, point)


def test_bev_contains_polyline_point_false():
    polyline = Polyline([Point(x,x,x) for x in range(5)])
    assert not bird_eye_view_contains(polyline, Point(1.2,1.3,-5))


@pytest_benchmark
def test_bev_contains_polyline_polyline_true(benchmark):
    pl1 = Polyline([Point(x,x,x) for x in range(5)])
    pl2 = Polyline([Point(x,x,-x) for x in range(1,4)])
    assert benchmark(bird_eye_view_contains, pl1, pl2)


def test_bev_contains_polyline_polyline_false():
    pl1 = Polyline([Point(x,x,x) for x in range(5)])
    pl2 = Polyline([Point(x,x,-x) for x in range(1,4)] + [Point(2,1,0)])
    assert not bird_eye_view_contains(pl1, pl2)


@pytest_benchmark
def test_bev_contains_polygon_point_true(benchmark):
    polygon = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    point = Point(0,0,4)
    assert benchmark(bird_eye_view_contains, polygon, point)


def test_bev_contains_polygon_point_false():
    polygon = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    assert not bird_eye_view_contains(polygon, Point(3,0,4))


@pytest_benchmark
def test_bev_contains_polygon_box_true(benchmark):
    polygon = Polygon([Point(2,2), Point(-2,2), Point(-2,-2), Point(2,-2)])
    box = BoundingBox(Point(0,0,-3))
    assert benchmark(bird_eye_view_contains, polygon, box)


def test_bev_contains_polygon_box_false():
    polygon = Polygon([Point(2,2), Point(-2,2), Point(-2,-2), Point(2,-2)])
    assert not bird_eye_view_contains(polygon, BoundingBox(Point(0,1.5,-3)))


@pytest_benchmark
def test_bev_contains_polygon_polyline_true(benchmark):
    polygon = Polygon([Point(3,3), Point(-3,3), Point(-3,-3), Point(3,-3)])
    polyline = Polyline([Point(x,x,x) for x in range(2)])
    assert benchmark(bird_eye_view_contains, polygon, polyline)


def test_bev_contains_polygon_polyline_false():
    polygon = Polygon([Point(2,2), Point(-2,2), Point(-2,-2), Point(2,-2)])
    polyline = Polyline([Point(x,x,x) for x in range(5)])
    assert not bird_eye_view_contains(polygon, polyline)


@pytest_benchmark
def test_bev_contains_polygon_polygon_true(benchmark):
    pg1 = Polygon([Point(3,3), Point(-3,3), Point(-3,-3), Point(3,-3)])
    pg2 = Polygon([Point(2,2,3), Point(-2,2), Point(-2,-2), Point(2,-2)])
    assert benchmark(bird_eye_view_contains, pg1, pg2)


def test_bev_contains_polygon_polygon_false():
    pg1 = Polygon([Point(3,3), Point(-3,3), Point(-3,-3), Point(3,-3)])
    pg2 = Polygon([Point(2,2,3), Point(-2,2), Point(-2,-2), Point(2,-2)])
    assert not bird_eye_view_contains(pg2, pg1)



from openride import bird_eye_view_distance
from openride.core.bounding_box import BoundingBox
from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.polygon import Polygon
from openride.core.polyline import Polyline
from openride.core.rotation import Rotation
from openride.core.size import Size
from openride.test import get_pytest_benchmark
from openride.test.random_core_generator import get_random

from shapely import geometry

import pytest
import numpy as np



pytest_benchmark = get_pytest_benchmark(group="operations.bird_eye_view_distance")


@pytest_benchmark
def test_bev_distance_point_point(benchmark):
    p1 = Point(0,0,0)
    p2 = Point(1,1,1)
    assert benchmark(bird_eye_view_distance, p1, p2) == 2**0.5


def test_bev_distance_point_box_inside():
    p = Point(0,0,0)
    b = BoundingBox()
    assert bird_eye_view_distance(p,b) == 0.0


@pytest_benchmark
def test_bev_distance_point_box(benchmark):
    p = Point(2,0,0)
    b = BoundingBox()
    assert benchmark(bird_eye_view_distance, p, b) == 1.0


@pytest_benchmark
def test_bev_distance_point_line(benchmark):
    p = Point(2,2,0)
    l = Polyline([Point(x) for x in range(5)])
    assert benchmark(bird_eye_view_distance, p, l) == 2.0


@pytest_benchmark
def test_bev_distance_point_polygon(benchmark):
    p = Point(2,2,0)
    poly = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    assert benchmark(bird_eye_view_distance, p, poly) == 2**0.5


@pytest_benchmark
def test_bev_distance_box_box(benchmark):
    b1 = BoundingBox(Point(0,0,0), Rotation(0,0,0), Size(1,1,1))
    b2 = BoundingBox(Point(2*2**0.5+1,0,0), Rotation(0,0,np.pi/4), Size(1,1,1))
    assert benchmark(bird_eye_view_distance, b1, b2) == pytest.approx(2**0.5)


def test_bev_distance_box_box_overlap():
    b1 = BoundingBox()
    b2 = BoundingBox(Point(0.5,0.5))
    assert bird_eye_view_distance(b1, b2) == 0.0


@pytest_benchmark
def test_bev_distance_box_line(benchmark):
    b = BoundingBox(Point(0,0,0), Rotation(0,0,0), Size(1,1,1))
    l = Polyline([Point(x+2) for x in range(10)])
    assert benchmark(bird_eye_view_distance, b, l) == pytest.approx(1.0)


def test_bev_distance_box_line_overlap():
    b = BoundingBox(Point(0,0,0), Rotation(0,0,0), Size(1,1,1))
    l = Polyline([Point(x) for x in range(10)])
    assert pytest.approx(bird_eye_view_distance(b, l)) == 0.0


@pytest_benchmark
def test_bev_distance_box_polygon(benchmark):
    b = BoundingBox(Point(3,0,0), Rotation(0,0,0), Size(1,1,1))
    p = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    assert benchmark(bird_eye_view_distance, b, p) == pytest.approx(1.0)


def test_bev_distance_box_polygon_overlap():
    b = BoundingBox(Point(0,0,0), Rotation(0,0,0), Size(1,1,1))
    p = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    assert pytest.approx(bird_eye_view_distance(b, p)) == 0.0


@pytest_benchmark
def test_bev_distance_line_point(benchmark):
    p = Point(5.3,5,5)
    l = Polyline([Point(x) for x in range(10)])
    assert benchmark(bird_eye_view_distance, l, p) == pytest.approx(5.0)


def test_bev_distance_line_point_overlap():
    p = Point(5.3,0,5)
    l = Polyline([Point(x) for x in range(10)])
    assert pytest.approx(bird_eye_view_distance(l, p)) == 0.0


@pytest_benchmark
def test_bev_distance_line_line(benchmark):
    l1 = Polyline([Point(x) for x in range(10)])
    l2 = Polyline([Point(5.5, y+1) for y in range(10)])
    assert benchmark(bird_eye_view_distance, l1, l2) == 1.0


def test_bev_distance_line_line_overlap():
    l1 = Polyline([Point(x) for x in range(10)])
    l2 = Polyline([Point(5, y-5) for y in range(10)])
    assert bird_eye_view_distance(l1, l2) == 0.0


def test_bev_distance_line_polygon_overlap():
    l = Polyline([Point(x) for x in range(10)])
    p = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    assert bird_eye_view_distance(l, p) == 0.0


@pytest_benchmark
def test_bev_distance_polygon_polygon(benchmark):
    p1 = Polygon([Point(1,1), Point(-1,1), Point(-1,-1), Point(1,-1)])
    p2 = Polygon([Point(2,0), Point(3,-1), Point(4,0), Point(3,1)])
    assert benchmark(bird_eye_view_distance, p1, p2) == pytest.approx(1.0)


def test_bev_distance_line_line_same_as_shapely():
    for _ in range(10):
        l1 = get_random(Polyline)
        l2 = get_random(Polyline)
        assert pytest.approx(bird_eye_view_distance(l1, l2)) == l1.to_shapely().distance(l2.to_shapely())


def test_bev_distance_point_point_same_as_shapely():
    for _ in range(10):
        p1 = get_random(Point)
        p2 = get_random(Point)
        assert pytest.approx(bird_eye_view_distance(p1, p2)) == p1.to_shapely().distance(p2.to_shapely())


def test_bev_distance_box_point_same_as_shapely():
    for _ in range(10):
        p = get_random(Point)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(b, p)) == b.to_shapely().distance(p.to_shapely())


def test_bev_distance_box_box_same_as_shapely():
    for _ in range(10):
        b1 = get_random(BoundingBox)
        b2 = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(b1, b2)) == b1.to_shapely().distance(b2.to_shapely())


def test_bev_distance_box_line_same_as_shapely():
    for _ in range(10):
        l = get_random(Polyline)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(b, l)) == b.to_shapely().distance(l.to_shapely())


def test_bev_distance_line_point_same_as_shapely():
    for _ in range(10):
        p = get_random(Point)
        l = get_random(Polyline)
        assert pytest.approx(bird_eye_view_distance(l, p)) == l.to_shapely().distance(p.to_shapely())


def test_bev_distance_line_box_same_as_shapely():
    for _ in range(10):
        l = get_random(Polyline)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(l, b)) == l.to_shapely().distance(b.to_shapely())
        

def test_bev_distance_polygon_point_same_as_shapely():
    for _ in range(10):
        poly = get_random(Polygon)
        p = get_random(Point)
        assert pytest.approx(bird_eye_view_distance(poly, p)) == poly.to_shapely().distance(p.to_shapely())
        

def test_bev_distance_polygon_box_same_as_shapely():
    for _ in range(10):
        p = get_random(Polygon)
        b = get_random(BoundingBox)
        assert pytest.approx(bird_eye_view_distance(p, b)) == p.to_shapely().distance(b.to_shapely())


def test_bev_distance_polygon_line_same_as_shapely():
    for _ in range(10):
        p = get_random(Polygon)
        l = get_random(Polyline)
        assert pytest.approx(bird_eye_view_distance(p, l)) == p.to_shapely().distance(l.to_shapely())
       

def test_bev_distance_polygons_same_as_shapely():
    for _ in range(10):
        p1 = get_random(Polygon)
        p2 = get_random(Polygon)
        assert pytest.approx(bird_eye_view_distance(p1, p2)) == p1.to_shapely().distance(p2.to_shapely())


def test_default_geometry_distance_from_shapely():

    class MockGeometry(Geometry):
        def to_shapely(self):
            return geometry.Point([1,1,1])
        def transform(self, transform):
            pass
    
    p1 = Point(0,0,0)
    p2 = MockGeometry()
    assert bird_eye_view_distance(p1, p2) == 2**0.5



from openride.viewer.model_loader import ModelLoader
from openride.viewer.models import MODEL_ROTATIONS
from openride.core.rotation import Rotation

import numpy as np
import vtk
import vtk.util.numpy_support as vtk_np

from matplotlib import cm
from typing import List



def get_point_actor(x:float, y:float, z:float, color:tuple=(1,1,1), size:int=0.15, alpha:float=1, resolution:int=10) -> vtk.vtkActor:
    source = vtk.vtkSphereSource()
    source.SetThetaResolution(resolution)
    source.SetPhiResolution(resolution)
    source.SetCenter(x, y, z)
    source.SetRadius(size)
    return source_to_actor(source, color, alpha)


BOX_ACTORS_CACHE = {}
def get_box_actor(x:float, y:float, z:float, sx:float, sy:float, sz:float, roll:float, pitch:float, yaw:float, color:tuple=None, alpha:float=1) -> vtk.vtkActor:
    
    # Creating these actors is expansive, so we reuse them if an identical one has already been created.
    args = (color, alpha)
    if args not in BOX_ACTORS_CACHE:
        source = vtk.vtkCubeSource()
        BOX_ACTORS_CACHE[args] = source_to_actor(source, color, alpha)

    actor = vtk.vtkActor()
    actor.ShallowCopy(BOX_ACTORS_CACHE[args])

    actor.SetScale(2 * sx, 2 * sy, 2 * sz)
    rotate_actor(actor, roll, pitch, yaw)
    move_actor(actor, x, y, z)
    return actor


def get_line_actor(vertices:np.ndarray, color:tuple=(1,1,1), alpha:float=1, linewidth:int=2, dashed:bool=False) -> vtk.vtkActor:
    polydata = vtk.vtkPolyData()
    actor = polydata_to_actor(polydata, color, alpha)
    actor.GetProperty().SetLineWidth(linewidth)
    points = vtk.vtkPoints()
    points.SetData(vtk_np.numpy_to_vtk(vertices))
    polydata.SetPoints(points)
    
    lines = vtk.vtkCellArray()
    step = 1 if not dashed else 2
    for i in range(0, vertices.shape[0]-1, step):
        lines.InsertNextCell(2)
        lines.InsertCellPoint(i)
        lines.InsertCellPoint(i+1)
    polydata.SetLines(lines)
    return actor


def get_lines_actor(list_vertices:List[np.ndarray], color:tuple=(1,1,1), alpha:float=1, linewidth:int=2) -> vtk.vtkActor:
    polydata = vtk.vtkPolyData()
    actor = polydata_to_actor(polydata, color, alpha)
    actor.GetProperty().SetLineWidth(linewidth)
    points = vtk.vtkPoints()
    points.SetData(vtk_np.numpy_to_vtk(np.vstack(list_vertices)))
    polydata.SetPoints(points)

    lines = vtk.vtkCellArray()
    i = 0
    for vertices in list_vertices:
        for _ in range(0, vertices.shape[0]-1, 1):
            lines.InsertNextCell(2)
            lines.InsertCellPoint(i)
            lines.InsertCellPoint(i+1)
            i += 1
        i += 1
    polydata.SetLines(lines)
    return actor


def get_polygon_polydata(vertices:np.ndarray) -> vtk.vtkPolyData:

    points = vtk.vtkPoints()
    polygon = vtk.vtkPolygon()
    polygon.GetPointIds().SetNumberOfIds(vertices.shape[0])
    
    for i, p in enumerate(vertices):
        polygon.GetPointIds().SetId(i, i)
        points.InsertNextPoint(p[0], p[1], p[2]) 

    # Add the polygon to a list of polygons
    polygons = vtk.vtkCellArray()
    polygons.InsertNextCell(polygon)

    # Create a PolyData
    polygonPolyData = vtk.vtkPolyData()
    polygonPolyData.SetPoints(points)
    polygonPolyData.SetPolys(polygons)

    pdflt = vtk.vtkTriangleFilter()
    pdflt.SetInputData(polygonPolyData)
    pdflt.Update()
    polygonPolyData = pdflt.GetOutput()

    return polygonPolyData


def get_polygon_actor(vertices:np.ndarray, color:tuple=(1,1,1), alpha:float=1) -> vtk.vtkActor:
    return polydata_to_actor(get_polygon_polydata(vertices), color, alpha)


def get_polygons_actor(vertices_list:List[np.ndarray], color:tuple=(1,1,1), alpha:float=1) -> vtk.vtkActor:

    appendFilter = vtk.vtkAppendPolyData()
    for vertices in vertices_list:
        appendFilter.AddInputData(get_polygon_polydata(vertices))
    appendFilter.Update()

    pdflt = vtk.vtkTriangleFilter()
    pdflt.SetInputData(appendFilter.GetOutput())
    pdflt.Update()
    polygonPolyData = pdflt.GetOutput()

    return polydata_to_actor(polygonPolyData, color, alpha)


def get_point_cloud_actor(vertices:np.ndarray, colors:np.ndarray=None, color:tuple=None, colormap:str='jet', size:int=10, alpha:float=1, color_range:float=2**16):
    polydata = vtk.vtkPolyData()
    actor = polydata_to_actor(polydata, alpha=alpha)
    actor.GetProperty().SetPointSize(size)
    actor.GetProperty().SetRenderPointsAsSpheres(1)
    points = vtk.vtkPoints()
    cells = vtk.vtkCellArray()
    N = vertices.shape[0]
    points.SetData(vtk_np.numpy_to_vtk(vertices))
    cells_npy = np.vstack([np.ones(N, dtype = np.int64), np.arange(N, dtype = np.int64)]).T.flatten()
    cells.SetCells(N, vtk_np.numpy_to_vtkIdTypeArray(cells_npy))
    polydata.SetPoints(points)
    polydata.SetVerts(cells)

    # Colorize point cloud
    # FIXME: This is slow for large point clouds because of the loop. But a better solution is not trivial.
    if color is None:
        vtkcolors = vtk.vtkUnsignedCharArray()
        vtkcolors.SetNumberOfComponents(3)
        normalized_colors = np.log(1 + np.clip(colors, 0, np.inf)) / np.log(color_range)
        colors = getattr(cm, colormap)(normalized_colors)[:,:3]*255
        for i, p in enumerate(vertices):
            vtkcolors.InsertNextTuple(colors[i])
        polydata.GetCellData().SetScalars(vtkcolors)
    elif isinstance(color, tuple):
        actor.GetProperty().SetColor(*color)
    else:
        # (N,3) rgb color array
        vtkcolors = vtk.vtkUnsignedCharArray()
        vtkcolors.SetNumberOfComponents(3)
        for i, p in enumerate(vertices):
            vtkcolors.InsertNextTuple(color[i]*255)
        polydata.GetCellData().SetScalars(vtkcolors)

    return actor


# Creating these actors is expensive, so we initialize a bunch of them once and reuse them
TEXT_ACTORS_CACHE = [vtk.vtkCaptionActor2D() for _ in range(100)]
def get_text_actor(text:str, x:float, y:float, z:float, fontsize:int=16, _text_actor_index:int=None) -> vtk.vtkCaptionActor2D:

    if _text_actor_index is None:
        actor = vtk.vtkCaptionActor2D()
    else:
        # Because vtkCaptionActor2D are expensive to initialize, we reuse the same actors stored in a list that we expand if needed
        while _text_actor_index >= len(TEXT_ACTORS_CACHE):
            TEXT_ACTORS_CACHE.append(vtk.vtkCaptionActor2D())
        actor = TEXT_ACTORS_CACHE[_text_actor_index]

    actor.SetCaption(text)
    actor.SetAttachmentPoint(x, y, z)
    actor.BorderOff()
    actor.GetTextActor().SetTextScaleModeToNone()
    actor.GetCaptionTextProperty().SetFontSize(fontsize)
    actor.GetCaptionTextProperty().SetShadow(0)
    return actor


def get_cone_actor(x1:float, y1:float, z1:float, x2:float, y2:float, z2:float, color:tuple=(1,1,1), alpha:float=1, headsize:float=1):
    cone_source = vtk.vtkConeSource()
    cone_source.SetHeight(2*headsize)
    cone_source.SetRadius(headsize)
    cone_source.SetCenter(x2, y2, z2)
    cone_source.SetDirection(x2-x1, y2-y1, z2-z1)
    return source_to_actor(cone_source, color, alpha)


MODEL_ACTORS_CACHE = {}
def get_model_actor(loader:ModelLoader, file:str, number_polygons:int, x:float, y:float, z:float, sx:float, sy:float, sz:float, roll:float, pitch:float, yaw:float, color:tuple=(1,1,1), alpha:float=1) -> vtk.vtkActor:

    # Creating these actors is expansive, so we reuse them if an identical one has already been created.
    args = (file, color, alpha, number_polygons)
    if args not in MODEL_ACTORS_CACHE:
        polydata = loader(file, n_polygons=number_polygons)
        _actor = polydata_to_actor(polydata, color, alpha)
        rotation = MODEL_ROTATIONS.get(file.split('/')[-1], Rotation())
        rotate_actor(_actor, rotation.roll, rotation.pitch, rotation.yaw)
        MODEL_ACTORS_CACHE[args] = _actor

    actor = vtk.vtkActor()
    actor.ShallowCopy(MODEL_ACTORS_CACHE[args])

    x0,x1,y0,y1,z0,z1 = actor.GetBounds()
    actor.SetScale(2*sx/(x1-x0), 2*sy/(y1-y0), 2*sz/(z1-z0))
    rotate_actor(actor, roll, pitch, yaw)
    x0,x1,y0,y1,z0,z1 = actor.GetBounds()
    move_actor(actor, x - (x0 + x1)/2, y - (y0 + y1)/2, z - (z0 + z1)/2)
    
    return actor
            

def mapper_to_actor(mapper:vtk.vtkPolyDataMapper, color:tuple=(1,1,1), alpha:float=1.0) -> vtk.vtkActor:
    actor = vtk.vtkActor()
    actor.SetMapper(mapper)
    actor.GetProperty().SetColor(*color)
    actor.GetProperty().SetOpacity(alpha)
    return actor


def source_to_actor(source, color:tuple=(1,1,1), alpha:float=1.0) -> vtk.vtkActor:
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputConnection(source.GetOutputPort())
    return mapper_to_actor(mapper, color, alpha)


def polydata_to_actor(polydata, color:tuple=(1,1,1), alpha:float=1.0) -> vtk.vtkActor:
    mapper = vtk.vtkPolyDataMapper()
    mapper.SetInputData(polydata)
    return mapper_to_actor(mapper, color, alpha)


def array_to_texture(array:np.ndarray) -> vtk.vtkTexture:
    grid = vtk.vtkImageData()
    grid.SetDimensions(array.shape[0], array.shape[1], 1)
    vtkarr = vtk_np.numpy_to_vtk(array.swapaxes(0,1).reshape((-1, array.shape[-1]), order='F'))
    vtkarr.SetName('Image')
    grid.GetPointData().AddArray(vtkarr)
    grid.GetPointData().SetActiveScalars('Image')
    vtex = vtk.vtkTexture()
    vtex.SetColorModeToDirectScalars()
    vtex.SetInputDataObject(grid)
    return vtex


def move_actor(actor:vtk.vtkActor, x:float, y:float, z:float):
    actor.AddPosition(x, y, z)


def rotate_actor(actor:vtk.vtkActor, roll:float, pitch:float, yaw:float):
    actor.AddOrientation(np.rad2deg(roll), np.rad2deg(pitch), np.rad2deg(yaw))


def set_scale_actor(actor:vtk.vtkActor, sx:float, sy:float, sz:float):
    actor.SetScale(sx, sy, sz)


from openride.core.transform import Transform

from abc import ABC, abstractclassmethod
from typing import Any



class Collection(ABC):

    @abstractclassmethod
    def __len__(self) -> int:
        """Amount of objects in the collection"""


    @abstractclassmethod
    def append(self, obj:Any):
        """Add an object to the collection"""


    @abstractclassmethod
    def extend(self, collection:'Collection'):
        """Add another collection to this collection"""


    @abstractclassmethod
    def pop(self, index:int) -> Any:
        """Remove and return the object at index"""
        

    def __repr__(self):
        s = f'{self.__class__.__name__}: (N={len(self)}) \n'
        for i, obj in enumerate(self):
            s += f'{i}: {obj} \n'
        return s[:-2]


    @abstractclassmethod
    def __getitem__(self, index:int) -> Any:
        """Return the object at index"""


    @abstractclassmethod
    def transform(self, transform:Transform) -> 'Collection':        
        """Return a transformed copy of this collection"""



from typing import Tuple
import numba
import numpy as np



@numba.njit(cache=True)
def rotation_matrix(roll:float, pitch:float, yaw:float) -> np.ndarray:
    s1 = np.sin(roll)
    c1 = np.cos(roll)
    s2 = np.sin(pitch)
    c2 = np.cos(pitch)
    s3 = np.sin(yaw)
    c3 = np.cos(yaw)

    return np.array([
        [c2*c3, s1*s2*c3-c1*s3, s1*s3+c1*s2*c3],
        [c2*s3, c1*c3+s1*s2*s3, c1*s2*s3-s1*c3],
        [-s2, s1*c2, c1*c2],
    ])


@numba.njit(cache=True)
def euler_angles(matrix:np.ndarray) -> Tuple[float, float, float]:

    sy = np.sqrt(matrix[0,0] * matrix[0,0] +  matrix[1,0] * matrix[1,0])
    singular = sy < 1e-6

    if not singular:
        x = np.arctan2(matrix[2,1] , matrix[2,2])
        y = np.arctan2(-matrix[2,0], sy)
        z = np.arctan2(matrix[1,0], matrix[0,0])
    else:
        x = np.arctan2(-matrix[1,2], matrix[1,1])
        y = np.arctan2(-matrix[2,0], sy)
        z = 0

    return x, y, z


@numba.njit(cache=True)
def transform_point(x:float, y:float, z:float, matrix:np.ndarray) -> np.ndarray:
    return np.dot(np.array([x, y, z]) + matrix[:3,3], matrix[:3, :3])


@numba.njit(cache=True)
def transform_vertices(vertices:np.ndarray, matrix:np.ndarray) -> np.ndarray:
    return np.dot(vertices + matrix[:3,3], matrix[:3, :3])


from openride import Point, BoundingBox

import numpy as np



class Camera:

    def __init__(self, viewer):
        self.viewer = viewer
        self._camera = viewer.renderer.GetActiveCamera()
        self._camera.SetClippingRange(0.1, 100000)
        self._camera.SetViewUp(0,0,1)

        self.set_position(Point(0,0,0))
        self.set_focus(Point(1,0,0))


    def set_position(self, position:Point):
        self._camera.SetPosition(position.x, position.y, position.z)

    
    def set_focus(self, focus:Point):
        self._camera.SetFocalPoint(focus.x, focus.y, focus.z)


    def get_position(self) -> Point:
        return Point(*self._camera.GetPosition())

    
    def get_focus(self) -> Point:
        return Point(*self._camera.GetFocalPoint())


    def follow(self, bounding_box:BoundingBox, distance:float=15, height:float=3, spring:float=0.0):
        self.set_focus(bounding_box.position)

        x = -distance * np.cos(bounding_box.rotation.yaw)
        y = -distance * np.sin(bounding_box.rotation.yaw)
        z = height

        target_position = bounding_box.position + Point(x, y, z)
        self.set_position(spring * self.get_position() + (1-spring) * target_position)


    def is_in_front(self, point:Point) -> bool:
        position = self.get_position()
        focus = self.get_focus()
        v1 = np.array([point.x-position.x, point.y-position.y, point.z-position.z])
        v2 = np.array([focus.x-position.x, focus.y-position.y, focus.z-position.z])
        return np.dot(v1, v2) >= 0



from openride.core.numba.transforms import rotation_matrix, euler_angles

from dataclasses import dataclass

import numpy as np



@dataclass
class Rotation:

    roll:  float = 0.
    pitch: float = 0.
    yaw:   float = 0.


    def __post_init__(self):
        self.roll = self.roll % (2*np.pi)
        self.pitch = self.pitch % (2*np.pi)
        self.yaw = self.yaw % (2*np.pi)


    def get_matrix(self) -> np.ndarray:
        return rotation_matrix(self.roll, self.pitch, self.yaw)


    @classmethod
    def from_matrix(cls, matrix:np.ndarray) -> 'Rotation':
        return cls(*euler_angles(matrix))


    def __add__(self, other:'Rotation') -> 'Rotation':
        return Rotation(self.roll + other.roll, self.pitch + other.pitch, self.yaw + other.yaw)


    def __sub__(self, other:'Rotation') -> 'Rotation':
        return Rotation(self.roll - other.roll, self.pitch - other.pitch, self.yaw - other.yaw)



from openride import Point, BoundingBox, Polyline, Polygon, bird_eye_view_distance
from openride.core.numba.polygon_utils import polygon_contains_point, polygon_contains_all_vertices
from openride.core.numba.distances_bev import bev_distance_between_box_and_point

from multipledispatch import dispatch



@dispatch(BoundingBox, Point)
def bird_eye_view_contains(obj1:BoundingBox, obj2:Point) -> bool:
    return bev_distance_between_box_and_point(obj1.position.x, obj1.position.y, obj1.size.x, obj1.size.y, obj1.rotation.yaw, obj2.x, obj2.y) == 0.0


@dispatch(BoundingBox, BoundingBox)
def bird_eye_view_contains(obj1:BoundingBox, obj2:BoundingBox) -> bool:
    return polygon_contains_all_vertices(obj1.get_bird_eye_view_vertices(), obj2.get_bird_eye_view_vertices())


@dispatch(BoundingBox, Polyline)
def bird_eye_view_contains(obj1:BoundingBox, obj2:Polyline) -> bool:
    return polygon_contains_all_vertices(obj1.get_bird_eye_view_vertices(), obj2.vertices)


@dispatch(BoundingBox, Polygon)
def bird_eye_view_contains(obj1:BoundingBox, obj2:Polygon) -> bool:
    return polygon_contains_all_vertices(obj1.get_bird_eye_view_vertices(), obj2.vertices)
    

@dispatch(Polyline, Point)
def bird_eye_view_contains(obj1:Polyline, obj2:Point) -> bool:
    return bird_eye_view_distance(obj1, obj2) == 0.0


@dispatch(Polyline, Polyline)
def bird_eye_view_contains(obj1:Polyline, obj2:Polyline) -> bool:
    return all([bird_eye_view_contains(obj1, point) for point in obj2])


@dispatch(Polygon, Point)
def bird_eye_view_contains(obj1:Polygon, obj2:Point) -> bool:
    return polygon_contains_point(obj1.vertices, obj2.x, obj2.y)


@dispatch(Polygon, BoundingBox)
def bird_eye_view_contains(obj1:Polygon, obj2:BoundingBox) -> bool:
    return polygon_contains_all_vertices(obj1.vertices, obj2.get_bird_eye_view_vertices())


@dispatch(Polygon, Polyline)
def bird_eye_view_contains(obj1:Polygon, obj2:Polyline) -> bool:
    return polygon_contains_all_vertices(obj1.vertices, obj2.vertices)


@dispatch(Polygon, Polygon)
def bird_eye_view_contains(obj1:Polygon, obj2:Polygon) -> bool:
    return polygon_contains_all_vertices(obj1.vertices, obj2.vertices)



import vtk

from abc import ABC, abstractclassmethod



class BaseModelLoader(ABC):

    @abstractclassmethod
    def __call__(self, file:str, n_polygons:int=100000) -> vtk.vtkPolyData:
        """Loads files into vtk.vtkPolyData instances"""



class STLLoader(BaseModelLoader):

    def __init__(self):
        self._loaded_files = {}

    def __call__(self, file:str, n_polygons:int=100000) -> vtk.vtkPolyData:
        """Loads .stl files into vtk.vtkPolyData instances"""
        if file not in self._loaded_files:
            reader = vtk.vtkSTLReader()
            reader.SetFileName(file)
            reader.Update()
            polydata = reader.GetOutput()

            decimate = vtk.vtkQuadricDecimation()
            decimate.SetInputData(polydata)
            decimate.SetTargetReduction(1.0 - n_polygons / polydata.GetNumberOfPolys())
            decimate.Update()
    
            self._loaded_files[file] = decimate.GetOutput()

        return self._loaded_files[file]


class OBJLoader(BaseModelLoader):

    def __init__(self):
        self._loaded_files = {}

    def __call__(self, file:str, n_polygons:int=100000) -> vtk.vtkPolyData:
        """Loads .stl files into vtk.vtkPolyData instances"""
        if file not in self._loaded_files:
            reader = vtk.vtkOBJReader()
            reader.SetFileName(file)
            reader.Update()
            polydata = reader.GetOutput()

            decimate = vtk.vtkQuadricDecimation()
            decimate.SetInputData(polydata)
            decimate.SetTargetReduction(1.0 - n_polygons / polydata.GetNumberOfPolys())
            decimate.Update()
    
            self._loaded_files[file] = decimate.GetOutput()

        return self._loaded_files[file]


class ModelLoader(BaseModelLoader):

    def __init__(self):
        self.loaders = {
            'stl': STLLoader(),
            'obj': OBJLoader(),
        }

    def __call__(self, file:str, n_polygons:int=100000) -> vtk.vtkPolyData:
        loader = self.loaders.get(file.split('.')[-1])
        if loader is None:
            raise NotImplementedError
        else:
            return loader(file, n_polygons)


from openride.core.bounding_box import BoundingBox
from openride.core.point import Point
from openride.core.polygon import Polygon
from openride.core.polyline import Polyline
from openride.core.rotation import Rotation
from openride.core.size import Size

import numpy as np



def get_random(cls):

    if cls == Point:
        return Point(*np.random.random(3)*10)

    elif cls == BoundingBox:
        return BoundingBox(
            Point(*np.random.random(3)*10),
            Rotation(0,0,np.random.random()*2*np.pi),
            Size(*np.random.random(3)*10)
        )
        
    elif cls == Polyline:
        return Polyline(np.random.random((10,3)))

    elif cls == Polygon:
        right_side = [Point(x+np.random.random()*0.3, np.random.random()-5) for x in range(10)]
        left_side = [Point(10-x-np.random.random()*0.3, np.random.random()+5) for x in range(10)]
        return Polygon(right_side + left_side)


from openride.core.geometry import Geometry
from openride.core.numba.transforms import transform_point

from dataclasses import dataclass
from shapely import geometry
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from openride.core.transform import Transform



@dataclass
class Point(Geometry):
    
    x: float = 0.
    y: float = 0.
    z: float = 0.


    def to_shapely(self) -> geometry.Point:
        return geometry.Point([self.x, self.y, self.z])


    def transform(self, transform:'Transform') -> 'Point':
        return Point(*transform_point(self.x, self.y, self.z, transform.get_matrix()))


    def __add__(self, other:'Point') -> 'Point':
        return Point(self.x + other.x, self.y + other.y, self.z + other.z)

    
    def __sub__(self, other:'Point') -> 'Point':
        return Point(self.x - other.x, self.y - other.y, self.z - other.z)


    def __mul__(self, factor:float) -> 'Point':
        return Point(factor * self.x, factor * self.y, factor * self.z)
    def __rmul__(self, factor:float) -> 'Point':
        return self * factor


    def __repr__(self):
        return f'Point(x:{self.x:.2f}, y:{self.y:.2f}, z:{self.z:.2f})'
    


from openride.core.numba.transforms import transform_vertices, rotation_matrix

import numba
import numpy as np



@numba.njit(cache=True)
def bev_box_vertices(position_x, position_y, size_x, size_y, yaw) -> np.ndarray:
    vertices = np.zeros((4,3))
    vertices[:,0] = size_x
    vertices[:,1] = size_y
    vertices[2:,0] *= -1
    vertices[1:3,1] *= -1
    matrix4x4 = np.eye(4)
    matrix4x4[:3,:3] = rotation_matrix(0,0,-yaw)
    vertices = transform_vertices(vertices, matrix4x4)
    vertices[:,0] += position_x
    vertices[:,1] += position_y
    return vertices[:,:2]


@numba.njit(cache=True)
def box_vertices(position_x, position_y, position_z, size_x, size_y, size_z, roll, pitch, yaw) -> np.ndarray:
    vertices = np.zeros((8,3))
    vertices[:,0] = size_x
    vertices[:,1] = size_y
    vertices[:,2] = size_z
    vertices[:2,0] *= -1
    vertices[6:,0] *= -1
    vertices[4:,1] *= -1
    vertices[::2,2] *= -1
    matrix4x4 = np.eye(4)
    matrix4x4[:3,:3] = rotation_matrix(roll, pitch, yaw)
    vertices = transform_vertices(vertices, matrix4x4)
    vertices[:,0] += position_x
    vertices[:,1] += position_y
    vertices[:,2] += position_z
    return vertices


@numba.njit(cache=True)
def vertices_inside_box(vertices, box_vertices):
    """Returns indices of the vertices that are inside the box"""
    P = box_vertices
    u, v, w = P[1]-P[0], P[2]-P[0], P[6]-P[0]
    pu, pv, pw = np.dot(vertices, u), np.dot(vertices, v), np.dot(vertices, w)
    inside_box = np.where((pu <= np.dot(P[1], u)) & (pu >= np.dot(P[0], u))\
                        & (pv <= np.dot(P[2], v)) & (pv >= np.dot(P[0], v))\
                        & (pw <= np.dot(P[6], w)) & (pw >= np.dot(P[0], w))
                        )[0]
    return inside_box
    


from openride.core.geometry import Geometry
from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.size import Size
from openride.core.transform import Transform
from openride.core.numba.box_utils import bev_box_vertices, box_vertices

from dataclasses import dataclass
from shapely import geometry

import numpy as np



@dataclass
class BoundingBox(Geometry):

    position: Point = Point(0,0,0)
    rotation: Rotation = Rotation()
    size: Size = Size()


    def to_shapely(self) -> geometry.Polygon:
        return geometry.Polygon(self.get_bird_eye_view_vertices())


    def transform(self, transform:'Transform') -> 'BoundingBox':
        return BoundingBox(
            self.position.transform(transform), 
            self.rotation + transform.rotation,
            self.size,
        )


    def get_transform(self) -> Transform:
        return Transform(self.position, self.rotation)


    def get_bird_eye_view_vertices(self) -> np.ndarray:
        return bev_box_vertices(self.position.x, self.position.y, self.size.x, self.size.y, self.rotation.yaw)


    def get_vertices(self) -> np.ndarray:
        return box_vertices(
            self.position.x, self.position.y, self.position.z, 
            self.size.x, self.size.y, self.size.z, 
            self.rotation.roll, self.rotation.pitch, self.rotation.yaw,
        )



from openride.core.point import Point
from openride.core.rotation import Rotation

from dataclasses import dataclass

import numpy as np



@dataclass
class Transform:
    
    translation: Point = Point(0,0,0)
    rotation: Rotation = Rotation(0,0,0)


    def __post_init__(self):
        self._cached_matrix = None
        self._cached_inverse_matrix = None


    def get_matrix(self) -> np.ndarray:
        if self._cached_matrix is None:
            self._cached_matrix = np.eye(4)
            self._cached_matrix[:3,:3] = self.rotation.get_matrix()
            self._cached_matrix[:3,3] = [self.translation.x, self.translation.y, self.translation.z]
        return self._cached_matrix


    def get_inverse_matrix(self) -> np.ndarray:
        if self._cached_inverse_matrix is None:
            tf = self.get_matrix()
            self._cached_inverse_matrix = np.zeros_like(tf)
            self._cached_inverse_matrix[:3,:3] = tf[:3,:3].T 
            self._cached_inverse_matrix[:3,3] = np.dot(-tf[:3,:3].T, tf[:3, 3])
            self._cached_inverse_matrix[3,3] = 1
        return self._cached_inverse_matrix


    def inverse(self) -> 'Transform':
        return Transform.from_matrix(self.get_inverse_matrix())


    @classmethod
    def from_matrix(cls, matrix:np.ndarray) -> 'Transform':
        tf = cls(Point(*matrix[:3,3]), Rotation.from_matrix(matrix))
        tf._cached_matrix = matrix
        return tf



from openride.core.categories import Category, SubCategory
from openride.core.rotation import Rotation

from typing import List, Union

import glob
import os
import numpy as np

FILEPATH = os.path.dirname(os.path.abspath(__file__))



DEFAULT_MODELS = {
    Category.Vehicle: f"{FILEPATH}/cars/Aston_martin.stl",
    Category.Pedestrian: f"{FILEPATH}/pedestrians/default_male.stl",
    Category.Cyclist: f"{FILEPATH}/cyclists/bicycle2.stl",
    Category.TrafficCone: f"{FILEPATH}/traffic_cones/cone.stl",

    SubCategory.Car: f"{FILEPATH}/cars/Aston_martin.stl",
    SubCategory.Van: f"{FILEPATH}/vans/volkswagen_transporter.stl",
    SubCategory.Bus: f"{FILEPATH}/buses/ShortSchoolBus.stl",
    SubCategory.Truck: f"{FILEPATH}/trucks/Truck2.stl",
    SubCategory.Motorcycle: f"{FILEPATH}/motorcycles/kawaski.stl",
}



DIRECTORIES = {
    Category.Vehicle: f"{FILEPATH}/cars",
    Category.Pedestrian: f"{FILEPATH}/pedestrians",
    Category.Cyclist: f"{FILEPATH}/cyclists",
    Category.TrafficCone: f"{FILEPATH}/traffic_cones",

    SubCategory.Car: f"{FILEPATH}/cars",
    SubCategory.Van: f"{FILEPATH}/vans",
    SubCategory.Bus: f"{FILEPATH}/buses",
    SubCategory.Truck: f"{FILEPATH}/trucks",
    SubCategory.Motorcycle: f"{FILEPATH}/motorcycles",
}



def get_model_files(category:Union[Category, SubCategory]) -> List[str]:
    directory = DIRECTORIES.get(category)
    if not directory: return []
    return glob.glob(f"{directory}/*")



MODEL_ROTATIONS = {

    'jeep_renegate.stl': Rotation(yaw=np.pi/2),
    'Aston_martin.stl': Rotation(yaw=np.pi),
    'chevrolet_camaro_ss.stl': Rotation(yaw=np.pi/2),
    'chevrolet_corvette_zr1.stl': Rotation(yaw=np.pi/2),
    'dodge_challenger.stl': Rotation(yaw=np.pi/2),
    'mitsubishi_lancer_evolution.stl': Rotation(yaw=np.pi/2),
    'nissan_skyline_gtr.stl': Rotation(yaw=np.pi/2),
    'shelby_mustang_gt350r.stl': Rotation(yaw=np.pi/2),
    'tesla_model_3.stl': Rotation(yaw=np.pi/2),
    'volkswagen_golf_gti.stl': Rotation(yaw=np.pi/2),
    'chopper_reduziert.stl': Rotation(roll=np.pi/2),
    'ducati.stl': Rotation(yaw=np.pi/2),
    'kawaski.stl': Rotation(yaw=np.pi/2),

    'volkswagen_transporter.stl': Rotation(yaw=np.pi/2),
    'vw_transporter2.stl': Rotation(yaw=np.pi/2),

    'Fire_Truck.stl': Rotation(),
    'trashtruck.stl': Rotation(),
    'Truck2.stl': Rotation(),

    'ShortSchoolBus.stl': Rotation(),

    'default_male.stl': Rotation(roll=np.pi/4, pitch=0, yaw=np.pi/2),
    'sophie_walk.stl': Rotation(roll=np.pi/2, yaw=np.pi/2),
    'tyler_walk.stl': Rotation(roll=np.pi/2, yaw=np.pi/2),
    'woman_stand.stl': Rotation(yaw=np.pi/2),
    'default_female.stl': Rotation(yaw=np.pi/2),
    'vahlen_stand.stl': Rotation(yaw=np.pi/2),

    'bicycle.stl': Rotation(),
    'bicycle2.stl': Rotation(),

}


from openride.core.geometry import Geometry
from openride.core.numba.interpolation import linear_interpolation
from openride.core.numba.polyline_utils import distance_to_index, index_to_distance, resample_linear, tangent_angles
from openride.core.numba.polyline_utils import radius_of_curvature, vertices_to_distances
from openride.core.numba.transforms import transform_vertices
from openride.core.point import Point

import copy
import numpy as np

from dataclasses import dataclass
from shapely import geometry
from typing import Tuple, Union, List, TYPE_CHECKING

if TYPE_CHECKING:
    from openride.core.transform import Transform



@dataclass
class Polyline(Geometry):

    vertices: Union[np.ndarray, List[Point]] = np.empty((0,3))


    def __post_init__(self):

        if isinstance(self.vertices, list):
            points = copy.deepcopy(self.vertices)
            self.vertices = np.zeros((len(points), 3))
            for i, point in enumerate(points):
                self.vertices[i] = point.x, point.y, point.z
    

    def to_shapely(self) -> Union[geometry.LineString, geometry.Point]:
        if len(self) > 1: return geometry.LineString(self.vertices[:,:2])
        elif len(self) == 1: return geometry.Point(self.vertices[0,:2])


    def transform(self, transform:'Transform') -> 'Polyline':
        return Polyline(transform_vertices(self.vertices, transform.get_matrix()))


    def __len__(self) -> int:
        return self.vertices.shape[0]


    def append(self, point:Point) -> 'Polyline':
        return Polyline(vertices=np.vstack([self.vertices, np.array([point.x, point.y, point.z])]))


    def extent(self, polyline:'Polyline') -> 'Polyline':
        return Polyline(vertices=np.vstack([self.vertices, polyline.vertices]))


    def get_distances(self, bird_eye_view:bool=False) -> np.ndarray:
        return vertices_to_distances(self.vertices, bird_eye_view)


    def get_total_distance(self, bird_eye_view:bool=False) -> float:
        return self.get_distances(bird_eye_view)[-1]


    def index_to_distance(self, index:float) -> float:
        return index_to_distance(self.vertices, index)


    def distance_to_index(self, distance:float) -> float:
        return distance_to_index(self.vertices, distance)


    def __getitem__(self, key:Union[slice, float, int]) -> Union['Polyline', Point]:

        if isinstance(key, slice):

            if isinstance(key.start, float):
                start = self[key.start]
                return Polyline(np.vstack([np.array([[start.x, start.y, start.z]]), self[int(key.start+1):key.stop:key.step].vertices]))

            if isinstance(key.stop, float):
                stop = self[key.stop]
                return Polyline(np.vstack([self[key.start:int(key.stop+1):key.step].vertices, np.array([[stop.x, stop.y, stop.z]])]))
  
            return Polyline(self.vertices[key])

        if key % 1 == 0.0: key = int(key)

        if isinstance(key, (int, np.int64)):
            return Point(*self.vertices[key])

        elif isinstance(key, float):
            x = linear_interpolation(self.vertices[:,0], key)
            y = linear_interpolation(self.vertices[:,1], key)
            z = linear_interpolation(self.vertices[:,2], key)
            return Point(x, y, z)


    def get_tangent_angles(self) -> np.ndarray:
        return tangent_angles(self.vertices)


    def get_radius_of_curvature(self) -> np.ndarray:
        return radius_of_curvature(self.vertices)


    def resample(self, delta:float, return_indices:bool=False) -> Union['Polyline', Tuple['Polyline', np.ndarray]]:
        N = int(self.get_total_distance() / delta + 0.5) + 1
        resampled_vertices, indices = resample_linear(self.vertices, N)
        resampled = Polyline(resampled_vertices)
        if return_indices:
            return resampled, indices
        return resampled



import numba
import numpy as np



@numba.njit(cache=True)
def polygon_contains_point(vertices:np.ndarray, x:float, y:float) -> bool:
    n = len(vertices)
    inside = False
    p2x = 0.0
    p2y = 0.0
    xints = 0.0
    p1x,p1y = vertices[0,:2]
    for i in range(n+1):
        p2x,p2y = vertices[i % n, :2]
        if y > min(p1y,p2y):
            if y <= max(p1y,p2y):
                if x <= max(p1x,p2x):
                    if p1y != p2y:
                        xints = (y-p1y)*(p2x-p1x)/(p2y-p1y)+p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
        p1x,p1y = p2x,p2y

    return inside



@numba.njit(cache=True)
def polygon_contains_all_vertices(poly_vertices:np.ndarray, vertices:np.ndarray) -> bool:
    for v in vertices[:,:2]:
        if not polygon_contains_point(poly_vertices, v[0], v[1]):
            return False
    return True



from setuptools import setup, find_packages

with open("README.md", 'r') as file:
    long_description = file.read()

requirements = [
    'shapely',
    'numba',
    'numpy',
    'dataclasses',
    'vtk',
    'matplotlib',
]

setup(name = 'openride',
      version = '0.1.0',
      author = 'Jean-Luc Déziel',
      author_email = 'jluc1011@hotmail.com',
      url = 'https://gitlab.com/jldez/openride',
      description = '',
      long_description = long_description,
      long_description_content_type = 'text/markdown',
      packages = find_packages(),
      install_requires = requirements,
    )


import openride
import pytest



def test_init():
    s = openride.Size(1,2,3)
    assert s.x == 1
    assert s.y == 2
    assert s.z == 3


def test_post_init_fail():
    with pytest.raises(ValueError):
        openride.Size(-1)


from openride.core.collections.collection import Collection
from openride.core.bounding_box import BoundingBox
from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.size import Size
from openride.core.transform import Transform
from openride.core.numba.transforms import transform_vertices

from typing import List

import numpy as np



class BoundingBoxCollection(Collection):

    def __init__(self, bounding_boxes:List[BoundingBox]=[]):

        self.dtype = np.dtype([('position', 'f8', (3)), ('rotation', 'f8', (3)), ('size', 'f8', (3))])
        self._raw:np.ndarray = np.empty(0, self.dtype)

        for bounding_box in bounding_boxes:
            self.append(bounding_box)


    def __len__(self) -> int:
        return self._raw.shape[0]


    def append(self, bbox:BoundingBox):
        array = np.array((
            (bbox.position.x, bbox.position.y, bbox.position.z),
            (bbox.rotation.roll, bbox.rotation.pitch, bbox.rotation.yaw),
            (bbox.size.x, bbox.size.y, bbox.size.z),
        ), self.dtype)
        self._raw = np.hstack((self._raw, array))


    def extend(self, box_collection:'BoundingBoxCollection'):
        self._raw = np.hstack((self._raw, box_collection._raw))


    def pop(self, index:int) -> BoundingBox:
        bbox = self[index]
        self._raw = np.delete(self._raw, index, axis=0)
        return bbox


    def __getitem__(self, index:int) -> BoundingBox:
        return BoundingBox(
            position = Point(*self._raw[index]['position']),
            rotation = Rotation(*self._raw[index]['rotation']),
            size = Size(*self._raw[index]['size']),
        )


    def transform(self, transform:Transform) -> 'BoundingBoxCollection':        
        tmp = np.copy(self._raw)
        tmp['position'] = transform_vertices(self._raw['position'], transform.get_matrix())
        tmp['rotation'][:,0] = (tmp['rotation'][:,0] + transform.rotation.roll) % (2*np.pi)
        tmp['rotation'][:,1] = (tmp['rotation'][:,1] + transform.rotation.pitch) % (2*np.pi)
        tmp['rotation'][:,2] = (tmp['rotation'][:,2] + transform.rotation.yaw) % (2*np.pi)
        transformed = BoundingBoxCollection()
        transformed._raw = tmp
        return transformed


    def get_distances(self) -> np.ndarray:
        return (self._raw['position'][:,0] ** 2 + self._raw['position'][:,1] ** 2 + self._raw['position'][:,2] ** 2) ** 0.5


    def filter(self, indices:List[int]) -> 'BoundingBoxCollection':
        filtered = BoundingBoxCollection()
        filtered._raw = self._raw[indices]
        return filtered

    
    def filter_distance(self, min_distance:float=0.0, max_distance:float=100.0) -> 'BoundingBoxCollection':
        distances = self.get_distances()
        return self.filter(np.where((distances >= min_distance) & (distances <= max_distance)))



from openride.core.point import Point
from openride.core.rotation import Rotation
from openride.core.transform import Transform
from openride.test import get_pytest_benchmark

from shapely import geometry

import numpy as np
import pytest



pytest_benchmark = get_pytest_benchmark(group="core.point")


def test_point_init():
    p = Point(1,2,3)
    assert p.x == 1 and p.y == 2 and p.z == 3


def test_point_to_shapely():
    p = Point(1,2,3)
    ps = p.to_shapely()
    assert isinstance(ps, geometry.Point)


@pytest_benchmark
def test_transform_point(benchmark):
    p = Point(0,0,0)
    tf = Transform()
    p2 = benchmark(p.transform, tf)
    assert p == p2


def test_translate_point():
    p = Point(1,2,3)
    tf = Transform(Point(4,5,6))
    p2 = p.transform(tf)
    assert p2 == Point(5,7,9)


def test_rotate_x_point():
    p = Point(1,1,1)
    tf = Transform(rotation=Rotation(roll=np.pi/2))
    p2 = p.transform(tf)
    assert pytest.approx(p2.x) == 1
    assert pytest.approx(p2.y) == 1
    assert pytest.approx(p2.z) == -1


def test_rotate_y_point():
    p = Point(1,1,1)
    tf = Transform(rotation=Rotation(pitch=np.pi/2))
    p2 = p.transform(tf)
    assert pytest.approx(p2.x) == -1
    assert pytest.approx(p2.y) == 1
    assert pytest.approx(p2.z) == 1


def test_rotate_z_point():
    p = Point(1,1,1)
    tf = Transform(rotation=Rotation(yaw=np.pi/2))
    p2 = p.transform(tf)
    assert pytest.approx(p2.x) == 1
    assert pytest.approx(p2.y) == -1
    assert pytest.approx(p2.z) == 1


@pytest_benchmark
def test_add_points(benchmark):
    p1 = Point(1,2,3)
    p2 = Point(4,5,6)
    assert benchmark(p1.__add__, p2) == Point(5,7,9)


def test_substract_points():
    p1 = Point(1,2,3)
    p2 = Point(4,5,6)
    assert p1 - p2 == Point(-3,-3,-3)


def test_multiply_point():
    p = Point(1,2,3)
    assert p*2 == Point(2,4,6)


def test_reverse_multiply_point():
    p = Point(1,2,3)
    assert 2*p == Point(2,4,6)


def test_repr_point():
    assert isinstance(str(Point()), str)



import openride as ride
from openride.core.polyline import Polyline



def main():

    viewer = ride.Viewer()

    viewer.camera.set_position(ride.Point(-10,0,6))
    viewer.camera.set_focus(ride.Point(4,0,0))

    viewer.hud.toggle_fps_count()
    viewer.hud.toggle_cursor_position()

    viewer.draw_grid(delta=5, extent=50)
    viewer.draw_grid(delta=1, alpha=0.3, extent=50)

    point = ride.Point()

    bounding_box = ride.BoundingBox(
        position = ride.Point(5, 4, 1),
        rotation = ride.Rotation(0, 0, 1),
        size = ride.Size(2, 1, 1)
    )

    polyline = ride.Polyline([ride.Point(x + 2, -3) for x in range(10)])

    polygon = ride.Polygon([ride.Point(7,1), ride.Point(2,1), ride.Point(3,-1), ride.Point(7,-1, 0.3)])

    while True:
        viewer.draw_point(point, persist=False)
        viewer.labelize('allo', point)
        viewer.draw_bounding_box(bounding_box, color=(0,1,0), alpha = 0.75)
        viewer.draw_model(ride.Category.Vehicle, bounding_box)
        viewer.draw_polyline(polyline, color = (1,0,1), linewidth = 3)
        viewer.draw_polygon(polygon, color = (1,0,0), alpha = 0.4)
        viewer.update()


if __name__ == "__main__":
    main()



from openride.core.numba.box_utils import vertices_inside_box

import numpy as np
import numba

EPSILON = np.finfo(float).eps



@numba.njit(cache=True)
def distance_between_points_3d(x1, y1, z1, x2, y2, z2) -> float:
    return ((x1-x2)**2 + (y1-y2)**2 + (z1-z2)**2)**0.5



BOX_TRIANGLE_INDICES = np.array([[0,1,2], [1,2,3], [0,2,4], [0,4,6], [2,3,4], [3,4,5], [0,1,6], [1,6,7], [4,5,6], [5,6,7], [1,3,5], [1,5,7]])

@numba.njit(cache=True)
def distance_between_box_and_point_3d(box_vertices, bx, by, bz, px, py, pz) -> float:
    point = np.array((px, py, pz))
    box_center = np.array([bx, by, bz])
    if len(vertices_inside_box(np.reshape(point, (1,3)), box_vertices)) > 0: return 0.0
    
    for triangle in BOX_TRIANGLE_INDICES:
        _triangle = np.vstack((box_vertices[triangle[0]], box_vertices[triangle[1]], box_vertices[triangle[2]]))
        if ray_triangle_intersect(box_center, point, _triangle):
            return distance_triangle_point_3d(_triangle, px, py, pz)


@numba.njit(cache=True)
def distance_between_line_and_point_3d(vertices:np.ndarray, x:float, y:float, z:float) -> float:
    distances = np.zeros(vertices.shape[0]-1)
    p = np.array((x, y, z))
    for i in range(vertices.shape[0]-1):
        distances[i] = point_segment_distance_3d(p, vertices[i], vertices[i+1])
    return np.min(distances)


@numba.njit(cache=True)
def distance_between_lines_3d(vertices1:np.ndarray, vertices2:np.ndarray) -> float:
    distances = np.zeros((vertices1.shape[0]-1, vertices2.shape[0]-1))
    for i in range(vertices1.shape[0]-1):
        for j in range(vertices2.shape[0]-1):
            distances[i,j] = segments_distance_3d(vertices1[i], vertices1[i+1], vertices2[j], vertices2[j+1])
    return np.min(distances)


@numba.njit(cache=True)
def distance_between_line_and_box_3d(line_vertices:np.ndarray, box_vertices:np.ndarray) -> float:
    if len(vertices_inside_box(line_vertices, box_vertices)) > 0: return 0.0
    distances = np.zeros((12, line_vertices.shape[0]-1))
    for i, face in enumerate(BOX_TRIANGLE_INDICES):
        triangle = np.vstack((box_vertices[face[0]], box_vertices[face[1]], box_vertices[face[2]]))
        for j in range(line_vertices.shape[0]-1):
            p1 = line_vertices[j]
            direction = line_vertices[j+1] - p1
            distances[i,j] = segment_triangle_distance(p1, direction, triangle)
            if distances[i,j] == 0.0: return 0.0
    return np.min(distances)


@numba.njit(cache=True)
def distance_between_boxes_3d(vertices1:np.ndarray, vertices2:np.ndarray) -> float:
    if len(vertices_inside_box(vertices1, vertices2)) > 0: return 0.0
    if len(vertices_inside_box(vertices2, vertices1)) > 0: return 0.0
    distances = []
    for face in BOX_TRIANGLE_INDICES:
        triangle = np.vstack((vertices1[face[0]], vertices1[face[1]], vertices1[face[2]]))
        for p in vertices2:
            distances.append(distance_triangle_point_3d(triangle, p[0], p[1], p[2]))
    for face in BOX_TRIANGLE_INDICES:
        triangle = np.vstack((vertices2[face[0]], vertices2[face[1]], vertices2[face[2]]))
        for p in vertices1:
            distances.append(distance_triangle_point_3d(triangle, p[0], p[1], p[2]))

    box_segment_indices = [6,7, 7,1, 1,0, 0,6,       4,5, 5,3, 3,2, 2,4,      6,4, 7,5, 0,2, 1,3]
    for i in range(0, len(box_segment_indices)-1, 2):
        v1 = vertices1[box_segment_indices[i]]
        v2 = vertices1[box_segment_indices[i+1]]
        for j in range(0, len(box_segment_indices)-1, 2):
            v3 = vertices2[box_segment_indices[j]]
            v4 = vertices2[box_segment_indices[j+1]]
            distances.append(segments_distance_3d(v1, v2, v3, v4))

    return min(distances)


@numba.njit(cache=True)
def point_segment_distance_3d(p, a, b):

    # normalized tangent vector
    diff = b - a
    norm = (diff[0]**2 + diff[1]**2 + diff[2]**2)**.5
    d = np.divide(diff, norm)

    # signed parallel distance components
    s = np.dot(a - p, d)
    t = np.dot(p - b, d)

    # clamped parallel distance
    h = max([s, t, 0.0])

    # perpendicular distance component
    c = np.cross(p - a, d)

    return np.hypot(h, np.linalg.norm(c))



@numba.njit(cache=True)
def clip(a, min, max):
    if a < min: return min
    elif a > max: return max
    return a


@numba.njit(cache=True)
def segments_distance_3d(p11, p12, p21, p22):
    r = p21 - p11
    u = p12 - p11
    v = p22 - p21

    ru = np.dot(r, u)
    rv = np.dot(r, v)
    uu = np.dot(u, u)
    uv = np.dot(u, v)
    vv = np.dot(v, v)

    det = uu*vv - uv*uv

    if det < EPSILON*uu*vv:
        s = ru/uu
        t = 0
    else:
        s = clip((ru*vv - rv*uv)/det, 0, 1)
        t = clip((ru*uv - rv*uu)/det, 0, 1)

    S = clip((t*uv + ru)/uu, 0, 1)
    T = clip((s*uv - rv)/vv, 0, 1)

    A = p11 + S*u
    B = p21 + T*v
    C = B - A
    return (C[0]**2+C[1]**2+C[2]**2)**0.5


@numba.njit(cache=True)
def distance_triangle_point_3d(triangle, px, py, pz):
    """https://gist.github.com/joshuashaffer/99d58e4ccbd37ca5d96e"""
    
    B = triangle[0, :]
    E0 = triangle[1, :] - B
    E1 = triangle[2, :] - B
    D = B - np.array((px, py, pz))
    a = np.dot(E0, E0)
    b = np.dot(E0, E1)
    c = np.dot(E1, E1)
    d = np.dot(E0, D)
    e = np.dot(E1, D)
    f = np.dot(D, D)

    det = a * c - b * b
    s = b * e - c * d
    t = b * d - a * e

    # Terible tree of conditionals to determine in which region of the diagram
    # shown above the projection of the point into the triangle-plane lies.
    if (s + t) <= det:
        if s < 0.0:
            if t < 0.0:
                # region4
                if d < 0:
                    t = 0.0
                    if -d >= a:
                        s = 1.0
                        sqrdistance = a + 2.0 * d + f
                    else:
                        s = -d / a
                        sqrdistance = d * s + f
                else:
                    s = 0.0
                    if e >= 0.0:
                        t = 0.0
                        sqrdistance = f
                    else:
                        if -e >= c:
                            t = 1.0
                            sqrdistance = c + 2.0 * e + f
                        else:
                            t = -e / c
                            sqrdistance = e * t + f

                            # of region 4
            else:
                # region 3
                s = 0
                if e >= 0:
                    t = 0
                    sqrdistance = f
                else:
                    if -e >= c:
                        t = 1
                        sqrdistance = c + 2.0 * e + f
                    else:
                        t = -e / c
                        sqrdistance = e * t + f
                        # of region 3
        else:
            if t < 0:
                # region 5
                t = 0
                if d >= 0:
                    s = 0
                    sqrdistance = f
                else:
                    if -d >= a:
                        s = 1
                        sqrdistance = a + 2.0 * d + f;  # GF 20101013 fixed typo d*s ->2*d
                    else:
                        s = -d / a
                        sqrdistance = d * s + f
            else:
                # region 0
                invDet = 1.0 / det
                s = s * invDet
                t = t * invDet
                sqrdistance = s * (a * s + b * t + 2.0 * d) + t * (b * s + c * t + 2.0 * e) + f
    else:
        if s < 0.0:
            # region 2
            tmp0 = b + d
            tmp1 = c + e
            if tmp1 > tmp0:  # minimum on edge s+t=1
                numer = tmp1 - tmp0
                denom = a - 2.0 * b + c
                if numer >= denom:
                    s = 1.0
                    t = 0.0
                    sqrdistance = a + 2.0 * d + f;  # GF 20101014 fixed typo 2*b -> 2*d
                else:
                    s = numer / denom
                    t = 1 - s
                    sqrdistance = s * (a * s + b * t + 2 * d) + t * (b * s + c * t + 2 * e) + f

            else:  # minimum on edge s=0
                s = 0.0
                if tmp1 <= 0.0:
                    t = 1
                    sqrdistance = c + 2.0 * e + f
                else:
                    if e >= 0.0:
                        t = 0.0
                        sqrdistance = f
                    else:
                        t = -e / c
                        sqrdistance = e * t + f
                        # of region 2
        else:
            if t < 0.0:
                # region6
                tmp0 = b + e
                tmp1 = a + d
                if tmp1 > tmp0:
                    numer = tmp1 - tmp0
                    denom = a - 2.0 * b + c
                    if numer >= denom:
                        t = 1.0
                        s = 0
                        sqrdistance = c + 2.0 * e + f
                    else:
                        t = numer / denom
                        s = 1 - t
                        sqrdistance = s * (a * s + b * t + 2.0 * d) + t * (b * s + c * t + 2.0 * e) + f

                else:
                    t = 0.0
                    if tmp1 <= 0.0:
                        s = 1
                        sqrdistance = a + 2.0 * d + f
                    else:
                        if d >= 0.0:
                            s = 0.0
                            sqrdistance = f
                        else:
                            s = -d / a
                            sqrdistance = d * s + f
            else:
                # region 1
                numer = c + e - b - d
                if numer <= 0:
                    s = 0.0
                    t = 1.0
                    sqrdistance = c + 2.0 * e + f
                else:
                    denom = a - 2.0 * b + c
                    if numer >= denom:
                        s = 1.0
                        t = 0.0
                        sqrdistance = a + 2.0 * d + f
                    else:
                        s = numer / denom
                        t = 1 - s
                        sqrdistance = s * (a * s + b * t + 2.0 * d) + t * (b * s + c * t + 2.0 * e) + f

    # account for numerical round-off error
    if sqrdistance < 0:
        sqrdistance = 0

    dist = sqrdistance ** 0.5

    return dist


@numba.njit(cache=True)
def ray_triangle_intersect(ray_origin, ray_direction, triangle):
    """https://github.com/johnnovak/raytriangle-test/blob/master/python/perftest.py"""
    v0, v1, v2 = triangle[0], triangle[1], triangle[2]
    v0v1 = v1 - v0
    v0v2 = v2 - v0
    pvec = np.cross(ray_direction, v0v2)
    det = np.dot(v0v1, pvec)
    if det < 0.000001: return False
    invDet = 1.0 / det
    tvec = ray_origin - v0
    u = np.dot(tvec, pvec) * invDet
    if u < 0 or u > 1: return False
    qvec = np.cross(tvec, v0v1)
    v = np.dot(ray_direction, qvec) * invDet
    if v < 0 or u + v > 1: return False
    if np.dot(v0v2, qvec) * invDet < 0.000001: return False
    return True


@numba.njit(cache=True)
def segment_triangle_intersects(ray_origin, ray_direction, triangle):
    intersects1 = ray_triangle_intersect(ray_origin, ray_direction, triangle)
    intersects2 = ray_triangle_intersect((ray_origin+ray_direction), ray_direction*-1, triangle[::-1])
    return intersects1 and intersects2


@numba.njit(cache=True)
def segment_triangle_distance(ray_origin, ray_direction, triangle):
    if segment_triangle_intersects(ray_origin, ray_direction, triangle): return 0.0
    p1 = ray_origin
    p2 = ray_origin + ray_direction
    distances = []
    for i,j in ((0,1),(1,2),(2,0)):
        distances.append(segments_distance_3d(p1, p2, triangle[i], triangle[j]))
    return min(distances)



from openride import BoundingBox, BoundingBoxCollection, Point, Polyline, Polygon, PointCloud
from openride.core.categories import Category, SubCategory
from openride.viewer import vtk_utils
from openride.viewer.camera import Camera
from openride.viewer.interactor import Interactor
from openride.viewer.hud import Hud
from openride.viewer.model_loader import ModelLoader
from openride.viewer.models import DEFAULT_MODELS

from typing import Iterable, List, Tuple, Union

import vtk



class Viewer:

    def __init__(self, resolution=(1600,900), background:Tuple[float,float,float]=(0.05,0.05,0.1), mouse_camera_interactions:bool=True):

        self.renderer = vtk.vtkRenderer()
        self.renderWindow = vtk.vtkRenderWindow()
        self.renderWindow.AddRenderer(self.renderer)

        self.set_resolution(resolution)
        self.set_background(background)

        self.camera = Camera(self)
        self.interactor = Interactor(self, mouse_camera_interactions)
        self.hud = Hud(self, position=(50,50))

        self._model_loader = ModelLoader()

        self._actors_to_clear = []
        self._text_actor_index = 0


    def set_resolution(self, resolution:Tuple[int, int]):
        self.renderWindow.SetSize(resolution[0], resolution[1])


    def set_background(self, color=Tuple[float,float,float]):
        self.renderer.SetBackground(*color)
        

    def update(self):
        self.interactor.update()
        self.hud.update()
        self.renderWindow.Render()
        self.__clear_actors()


    def render_actor(self, actor:vtk.vtkActor, persist:bool=False):
        self.renderer.AddActor(actor)
        if not persist: self._actors_to_clear.append(actor)


    def draw_point(self, point:Point, color:tuple=(1,1,1), size:int=0.15, alpha:float=1, resolution:int=10, persist:bool=False):
        actor = vtk_utils.get_point_actor(point.x, point.y, point.z, color, size, alpha, resolution)
        self.render_actor(actor, persist)


    def draw_point_cloud(self, pcloud:PointCloud, color:tuple=None, colors:Iterable[float]=None, colormap:str='jet', color_range:float=2**16, size:int=10, alpha:float=1, persist:bool=False):
        if color is None and colors is None: color = (1,1,1)
        actor = vtk_utils.get_point_cloud_actor(pcloud.get_point_cloud(), colors, color, colormap, size, alpha, color_range)
        self.render_actor(actor, persist)


    def draw_bounding_box(self, bounding_box:Union[BoundingBox, BoundingBoxCollection], color:tuple=(1,1,1), alpha:float=1, edge_color:tuple=None, persist:bool=False):

        if isinstance(bounding_box, BoundingBoxCollection):
            [self.draw_bounding_box(box, color, alpha, edge_color, persist) for box in bounding_box]
            return

        actor = vtk_utils.get_box_actor(
            bounding_box.position.x, bounding_box.position.y, bounding_box.position.z, 
            bounding_box.size.x, bounding_box.size.y, bounding_box.size.z,
            bounding_box.rotation.roll, bounding_box.rotation.pitch, bounding_box.rotation.yaw,
            color, alpha
        )

        if edge_color is not None:
            actor.GetProperty().SetEdgeVisibility(1)
            actor.GetProperty().SetLineWidth(10)
            actor.GetProperty().SetEdgeColor(*edge_color)

        self.render_actor(actor, persist)


    def draw_model(self, model:Union[str, Category, SubCategory], bounding_box:BoundingBox, number_polygons:int=100000, color:tuple=(1,1,1), alpha:float=1, persist:bool=False):
        """TODO"""
        p = bounding_box.position
        s = bounding_box.size
        r = bounding_box.rotation
        if isinstance(model, (Category, SubCategory)):
            file = DEFAULT_MODELS.get(model)
        else:
            file = model
        if not file: return
        actor = vtk_utils.get_model_actor(self._model_loader, file, number_polygons, p.x, p.y, p.z, s.x, s.y, s.z, r.roll, r.pitch, r.yaw, color, alpha)
        self.render_actor(actor, persist)


    def draw_polyline(self, line:Polyline, color:tuple=(1,1,1), alpha:float=1, linewidth:int=2, persist:bool=False):
        actor = vtk_utils.get_line_actor(line.vertices, color, alpha, linewidth)
        self.render_actor(actor, persist)


    def draw_polylines(self, lines:List[Polyline], color:tuple=(1,1,1), alpha:float=1, linewidth:int=2, persist:bool=False):
        actor = vtk_utils.get_lines_actor([line.vertices for line in lines], color=color, alpha=alpha, linewidth=linewidth)
        self.render_actor(actor, persist)


    def draw_polygon(self, polygon:Polygon, color:tuple=(1,1,1), alpha:float=1, persist:bool=False):
        actor = vtk_utils.get_polygon_actor(polygon.vertices, color=color, alpha=alpha)
        self.render_actor(actor, persist)


    def labelize(self, label:str, pos:Point, fontsize:int=16, persist:bool=False):
        if not self.camera.is_in_front(pos): return
        actor = vtk_utils.get_text_actor(label, pos.x, pos.y, pos.z, fontsize, self._text_actor_index)
        self.render_actor(actor, persist)
        self._text_actor_index += 1


    def draw_arrow(self, start:Point, end:Point, color:tuple=(1,1,1), alpha:float=1, linewidth:int=10, headsize:float=1, persist:bool=False):
        self.draw_line(Polyline([start, end]), color=color, alpha=alpha, linewidth=linewidth)
        actor = vtk_utils.get_cone_actor(
            start.x, start.y, start.z,
            end.x, end.y, end.z, 
            color, alpha, headsize,
        )
        self.render_actor(actor, persist)


    def draw_grid(self, origin:Point=Point(), delta:float=5, extent:float=200, alpha:float=1.0, persist:bool=True):
        lines = []

        for s in [0,1]:
            for i in range(int(2*extent/delta)+1):
                points = []
                if s == 0:
                    points.append(Point(-extent, -extent + i * delta) + origin)
                    points.append(Point( extent, -extent + i * delta) + origin)
                else:
                    points.append(Point(-extent + i * delta, -extent) + origin)
                    points.append(Point(-extent + i * delta,  extent) + origin)
                lines.append(Polyline(points))

        self.draw_polylines(lines, color=(0.3,0.3,0.3), alpha=alpha, linewidth=1, persist=persist)


    def __clear_actors(self):
        for actor in self._actors_to_clear:
            self.renderer.RemoveActor(actor)
        self._actors_to_clear = []
        self._text_actor_index = 0


    def close(self):
        del self.renderWindow
        self.interactor._interactor.InvokeEvent("ExitEvent")



import pytest

PYTEST_BENCHMARK_PARAMS = {
    'min_time':0.0001,
    'max_time':0.025,
    'min_rounds':5,
    'warmup':True,
    'warmup_iterations':1,
}


def get_pytest_benchmark(group:str):
    return pytest.mark.benchmark(group=group, **PYTEST_BENCHMARK_PARAMS)